//
//  contourDetector.java
//  detectionContours
//
//  Created by Thomas Jannaud on 23/12/09.
//  Copyright 2009 Jannaud's Technical Company. All rights reserved.
//

import java.util.HashSet;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Collection;
import java.awt.Rectangle;

public class contourDetector extends Thread {
	private myGreyPicture image;
	private int width, height = 0;
	public boolean setGrey = true;
	
	public static final int GRADIENT_X = 0, GRADIENT_Y = 1, GRADIENT_XY = 2, GRADIENT_SOBEL = 3, GRADIENT_CANNY = 4;
	public static final int LAPLACIEN = 5, GRAPHE_MIN_CUT = 6, GRAPHE_MIN_CUT_ARRET_CUT = 7, GRAPHE_MIN_CUT_ARRET_RATIO = 8;
	private int detectionMode = -1, thickness = 1;
	private Object selection;
	private boolean notFinished = true;
	private showMe publisher;
		
	public contourDetector(myGreyPicture im, int detectionMode, boolean isGrey, showMe shower, int thickness) {
		image = im;
		setGrey = isGrey;
		this.detectionMode = detectionMode;
		this.width = im.getWidth();
		this.height = im.getHeight();
		publisher = shower;
		this.thickness = thickness;
	}
	
	public void setSelection(Object obj) {selection = obj;}
	
	public void finished() {
		notFinished = false;
		if (publisher != null)
			publisher.finished();
	}
	
	public void run() {
		notFinished = true;
		if (detectionMode == GRADIENT_X)
			detect_grad(1.f, 0.f, 0.15f, 0.2f);
		else if (detectionMode == GRADIENT_Y)
			detect_grad(0.f, 1.f, 0.15f, 0.2f);
		else if (detectionMode == GRADIENT_XY)
			detect_grad(0.5f, 0.5f, 0.1f, 0.15f);
		else if (detectionMode == GRADIENT_SOBEL)
			detect_sobel(0.07f, 0.1f);
		else if (detectionMode == LAPLACIEN)
			detect_laplacien(0.07f, 0.1f);
		else if (detectionMode == GRADIENT_CANNY && selection != null) {
			Rectangle r = (Rectangle)(selection);
			detect_canny(r.x, r.y, r.width, r.height);
		}
		else if (detectionMode == GRADIENT_CANNY)
			detect_canny(.05f, .07f);
		else if (detectionMode == GRAPHE_MIN_CUT && selection != null)
			detect_minCutContour1((Collection<myPoint>) (selection));
		else if (detectionMode == GRAPHE_MIN_CUT_ARRET_RATIO && selection != null)
			detect_minCutContour2((Collection<myPoint>) (selection));
		else if (detectionMode == GRAPHE_MIN_CUT_ARRET_CUT && selection != null)
			detect_minCutContour3((Collection<myPoint>) (selection));

		if (publisher != null)
			publisher.finished();
	}
	
	public void detect_grad(float weight_x, float weight_y, float low_threshold, float high_threshold) {
		float[] newImage = new float[width * height];
		float[][] mask_x = {{-1.0f, 0.0f, 1.0f}};
		float[][] mask_y = {{-1.0f}, {0.0f}, {1.0f}};
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++) {
				float gradX = apply_mask(mask_x, i, j);
				float gradY = apply_mask(mask_y, i, j);
				float grad = weight_x * Math.abs(gradX) + weight_y * Math.abs(gradY);
				newImage[i + j * width] = grad;
			}
		float[] newImage2 = apply_hysteresis_threshold(newImage, low_threshold, high_threshold, width, height);
		publish(newImage2, width, height);
	}
	
	public void detect_laplacien(float low_threshold, float high_threshold) {
		float[] newImage = new float[width * height];
		float[][] mask = {{-1.0f, -1.0f, -1.0f}, {-1.0f, 8.0f, -1.0f}, {-1.0f, -1.0f, -1.0f}};
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++) {
				float laplacien = apply_mask(mask, i, j);
				newImage[i + j * width] = laplacien;
			}
		float[] newImage2 = apply_hysteresis_threshold(newImage, low_threshold, high_threshold, width, height);
		publish(newImage2, width, height);
	}
	
	public void detect_sobel(float low_threshold, float high_threshold) {
		float[] newImage = new float[width * height];
		float[][] mask_x = {
			{-1.0f, 0.0f, 1.0f},
			{-2.0f, 0.0f, 2.0f},
			{-1.0f, 0.0f, 1.0f}};
		float[][] mask_y = {
			{-1.f, -2.f, -1.f},
			{0.0f, 0.0f, 0.0f},
			{1.0f, 2.0f, 1.0f}};
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++) {
				float gradX = apply_mask(mask_x, i, j);
				float gradY = apply_mask(mask_y, i, j);
				float grad = 1.f / 16.0f * (Math.abs(gradX) + Math.abs(gradY));
				newImage[i + j * width] = grad;
			}
		float[] newImage2 = apply_hysteresis_threshold(newImage, low_threshold, high_threshold, width, height);
		publish(newImage2, width, height);
	}
	
	public void detect_canny(float low_threshold, float high_threshold) {
		float[] newImage = new float[width * height];
		float[] newImage2 = new float[width * height];
		
		// gaussian smoothing
		float coef = 1.0f / 159.0f;
		float[][] gaussienne_5x5 = {
			{2.0f * coef, 4.0f * coef, 5.0f * coef, 4.0f * coef, 2.0f * coef}, 
			{4.0f * coef, 9.0f * coef, 12.0f * coef, 9.0f * coef, 4.0f * coef}, 
			{5.0f * coef, 12.0f * coef, 15.0f * coef, 12.0f * coef, 5.0f * coef}, 
			{4.0f * coef, 9.0f * coef, 12.0f * coef, 9.0f * coef, 4.0f * coef}, 
			{2.0f * coef, 4.0f * coef, 5.0f * coef, 4.0f * coef, 2.0f * coef}};
		for(int i = 2; i < width - 2; i++)
			for(int j = 2; j < height - 2; j++)
				newImage[i + j * width] = apply_mask(gaussienne_5x5, i, j);
		
		// gradient
		float[][] mask_x = {{-1.0f, 0.0f, 1.0f}};
		float[][] mask_y = {{-1.0f}, {0.0f}, {1.0f}};
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++) {
				float gradX = apply_mask(mask_x, i, j, newImage);
				float gradY = apply_mask(mask_y, i, j, newImage);
				newImage2[i + j * width] = 0.5f * (Math.abs(gradX) + Math.abs(gradY));
			}
		
		apply_hysteresis_threshold(newImage2, newImage, low_threshold, high_threshold, width, height);
		publish(newImage, width, height);
	}
	
	public void detect_canny(int minX, int minY, int larg, int haut) {
		// Canny but on a zone
		// we find the threshold so that a certain percentage only is marked as a contour.
		
		float[] newImage = new float[width * height];
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				newImage[i + j * width] = image.greyComponent(i, j);
		
		// gaussian smoothing
		float coef = 1.0f / 159.0f;
		float[][] gaussienne_5x5 = {
			{2.0f * coef, 4.0f * coef, 5.0f * coef, 4.0f * coef, 2.0f * coef}, 
			{4.0f * coef, 9.0f * coef, 12.0f * coef, 9.0f * coef, 4.0f * coef}, 
			{5.0f * coef, 12.0f * coef, 15.0f * coef, 12.0f * coef, 5.0f * coef}, 
			{4.0f * coef, 9.0f * coef, 12.0f * coef, 9.0f * coef, 4.0f * coef}, 
			{2.0f * coef, 4.0f * coef, 5.0f * coef, 4.0f * coef, 2.0f * coef}};
		for(int i = -2; i <= larg + 1; i++)
			for(int j = -2; j <= haut + 1; j++)
				newImage[minX + i + (minY + j) * width] = apply_mask(gaussienne_5x5, minX + i, minY + j);
		
		// gradient
		float[] newImage2 = new float[larg * haut];
		float[][] mask_x = {{-1.0f, 0.0f, 1.0f}};
		float[][] mask_y = {{-1.0f}, {0.0f}, {1.0f}};
		for(int i = 0; i < larg; i++)
			for(int j = 0; j < haut; j++) {
				float gradX = apply_mask(mask_x, minX + i, minY + j, newImage);
				float gradY = apply_mask(mask_y, minX + i, minY + j, newImage);
				newImage2[i + j * larg] = 0.5f * (Math.abs(gradX) + Math.abs(gradY));
			}
		
		// compute thresholds
		float percent_target = 0.20f;
		float high_threshold = quantile(newImage2, (int)((1.f - percent_target) * (float)(larg * haut)));
		float low_threshold = 0.7f * high_threshold;
		
		float[] newImage3 = apply_hysteresis_threshold(newImage2, low_threshold, high_threshold, larg, haut);
		
		// copy zone in the big image
		for(int i = 0; i < larg; i++)
			for(int j = 0; j < haut; j++)
				newImage[minX + i + (minY + j) * width] = newImage3[i + j * larg];
		
		if (setGrey)
			image.setGrey(newImage);
		else
			image.setRed(image_to_collection(newImage3, width, height));
	}
	
	public void detect_minCutContour1(Collection<myPoint> contour) {
		final int border_thickness = thickness;
		int lastSize = contour.size(), lastVal = 100000;
		Collection<myPoint> contours_before = copy(contour);
		Collection<myPoint> contours = copy(contour);
		int step = 0;
		while(notFinished) {
			step++;
			System.out.println("step " + step + ", cutVal " + lastVal + " ratio cutVal / size " + (double)(lastVal)/(double)(lastSize));
			dilate_contours(contours, border_thickness); 
			partition(contours);	
			myGraph g = myGraph.create_graphe(contours, image, width, height);
			pair<Collection<myPoint>, Integer> p = g.minCut();
			if (p.key.size() == 0)
				break;
			contours = p.key;
			if (contours.equals(contours_before))
				break;
			lastVal = p.value;
			lastSize = contours.size();
			image.unsetRed(contours_before);
			contours_before = copy(contours);
			image.setRed(contours_before);
		}
		System.out.println("finished");
		if (publisher != null)
			publisher.finished();
		return;
	}


	public void detect_minCutContour2(Collection<myPoint> contour) {
		final int border_thickness = thickness;
		int lastSize = contour.size(), lastVal = 100000;
		Collection<myPoint> contours_before = copy(contour);
		Collection<myPoint> contours = copy(contour);
		int step = 0;
		while(notFinished) {
			step++;
			System.out.println("step " + step + ", cutVal " + lastVal + " ratio cutVal / size " + (double)(lastVal)/(double)(lastSize));
			dilate_contours(contours, border_thickness); 
			partition(contours);	
			myGraph g = myGraph.create_graphe(contours, image, width, height);
			pair<Collection<myPoint>, Integer> p = g.minCut();
			if (p.key.size() == 0)
				break;
			double coef = (double)(lastVal) / (double)(lastSize);
			double newcoef = (double)(p.value) / (double)(p.key.size());
			if (newcoef >= coef)
				break;
			contours = p.key;
			lastVal = p.value;
			lastSize = contours.size();
			image.unsetRed(contours_before);
			contours_before = copy(contours);
			image.setRed(contours_before);
		}
		System.out.println("finished");
		if (publisher != null)
			publisher.finished();
		return;
	}

	public void detect_minCutContour3(Collection<myPoint> contour) {
		final int border_thickness = thickness;
		int lastSize = contour.size(), lastVal = 100000;
		Collection<myPoint> contours_before = copy(contour);
		Collection<myPoint> contours = copy(contour);
		int step = 0;
		while(notFinished) {
			step++;
			System.out.println("step " + step + ", cutVal " + lastVal + " ratio cutVal / size " + (double)(lastVal)/(double)(lastSize));
			dilate_contours(contours, border_thickness); 
			partition(contours);	
			myGraph g = myGraph.create_graphe(contours, image, width, height);
			pair<Collection<myPoint>, Integer> p = g.minCut();
			if (p.key.size() == 0)
				break;
			if (p.value >= lastVal)
				break;
			contours = p.key;
			lastVal = p.value;
			lastSize = contours.size();
			image.unsetRed(contours_before);
			contours_before = copy(contours);
			image.setRed(contours_before);
		}
		System.out.println("finished");
		if (publisher != null)
			publisher.finished();
		return;
	}


	// -------------------
	// auxiliary functions
	// -------------------
	
	private Collection<myPoint> copy(Collection<myPoint> c) {
		LinkedList l = new LinkedList();
		for(myPoint p : c)
			l.add(new myPoint(p));
		return l;
	}
	
	private float apply_mask(float[][] matrix, int x, int y) {
		float ret = 0.0f;
		final int ny = matrix.length, nx = matrix[0].length;
		final int ax = nx / 2, ay = ny / 2;
		for(int i = 0; i < nx; i++)
			for(int j = 0; j < ny; j++)
				ret += matrix[j][i] * image.greyComponent(x - ax + i, y - ay + j);
		return ret;
	}
	
	private float apply_mask(float[][] matrix, int x, int y, float[] im) {
		float ret = 0.0f;
		final int ny = matrix.length, nx=matrix[0].length;
		final int ax = nx / 2, ay = ny / 2;
		for(int i = 0; i < nx; i++)
			for(int j = 0; j < ny; j++)
				ret += matrix[j][i] * im[x - ax + i + (y - ay + j) * width];
		return ret;
	}
	
	private void publish(float[] newImage, int width, int height) {
		if (setGrey)
			image.setGrey(newImage);
		else
			image.setRed(image_to_collection(newImage, width, height));	
	}
	
	private Collection<myPoint> image_to_collection(float[] values, int larg, int haut) {
		HashSet<myPoint> c = new HashSet<myPoint>();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				if (values[i + j * width] > 0.99f)
					c.add(new myPoint(i, j));
		return c;
	}
	
	private float[] apply_hysteresis_threshold(float[] values, float low_threshold, float high_threshold, int larg, int haut) {
		float[] newImage = new float[larg * haut];
		apply_hysteresis_threshold(values, newImage, low_threshold, high_threshold, larg, haut);
		return newImage;
	}
	
	private void apply_hysteresis_threshold(float[] values, float[] returnValues, float low_threshold, float high_threshold, int larg, int haut) {
		// low threshold
		for(int i = 0; i < larg; i++)
			for(int j = 0; j < haut; j++)
				if (values[i + j * larg] < high_threshold)
					returnValues[i + j * larg] = 0.0f;
		
		// high threshold neighbours
		for(int i = 1; i < larg - 1; i++)
			for(int j = 1; j < haut - 1; j++)
				if (values[i + j * larg] >= high_threshold) {
					for(int ii = -1; ii <= 1; ii++)
						for(int jj = -1; jj <= 1; jj++)
							if (values[i + ii + (j + jj) * larg] >= low_threshold)
								returnValues[i + ii + (j + jj) * larg] = 1.0f;
				}
	}
	
	private float quantile(float[] t, int q) {
		// find the i_eme element of the tab, in ascending order, in O(n)
		final int n = t.length;
		float[] tab = new float[n];
		System.arraycopy(t, 0, tab, 0, n);
		int begin = 0, end = n;
		q = Math.max(Math.min(end - 1, q), 0);
		while(end > begin + 1) {
			int pbeg = begin;
			int pend = end;
			final float c = tab[begin];
			for(int i = 0; i < (end - begin - 1); i++) {
				float x = tab[pbeg + 1];
				if (x <= c) {
					swap(tab, pbeg, pbeg + 1);
					pbeg++;
				}
				else {
					pend--;
					swap(tab, pend, pbeg+1);
				}
			}
			if (q <= (pbeg - begin))
				end = pbeg + 1;
			else {
				q -= (pbeg - begin + 1);
				begin = pbeg + 1;
			}
		}
		return tab[begin];
	}
	
	private void swap(float[] t, int i, int j) {
		final float temp = t[i];
		t[i] = t[j];
		t[j] = temp;
	}
	
	// ---------------
	// graph functions
	// ---------------
	
	private void dilate_contours(Collection<myPoint> contours, int thickness) {
		// add a belt to contours. Points had type CONTOURS, they will have type CONTOURS or BELT
		LinkedList< pair<myPoint, Integer> > queue = new LinkedList< pair<myPoint, Integer> >();
		HashMap<myPoint, Integer> done = new HashMap<myPoint, Integer>();
		
		for(myPoint p : contours)
			queue.add(new pair(p, new Integer(thickness)));		
		
		myPoint.setComparator(myPoint.COMPARE_POINT);
		while (!queue.isEmpty()) {
			pair<myPoint, Integer> pp = queue.removeFirst();
			myPoint p = pp.key;
			int count = pp.value;
			done.put(p, new Integer(count));
			if (count > 0) {
				// 8 connexity neighbours
				for(int i = -1; i <= 1; i++)
					for(int j = -1; j <= 1; j++) {
						myPoint q = new myPoint(p.x + i, p.y + j, myPoint.BELT);
						if (!done.containsKey(q) || done.get(q) < (count - 1))
							queue.addLast(new pair(q, count - 1));
					}
			}
		}
		contours.addAll(done.keySet());
	}
	
	
	public static void partition(Collection<myPoint> belt) {
		// Points had type CONTOURS or BELT, they will have type CONTOURS, INSIDE or OUTSIDE
		// find the leftmost point : it is on the outside belt
		int xmin = 10000;
		myPoint pxMin = null;
		for(myPoint p : belt)
			if (p.x < xmin) {
				pxMin = p;
				xmin = p.x;
			}
		pxMin.type = myPoint.OUTSIDE_BELT;
		
		// all the points connected to it in the belt are outside on the outside
		// delimiter : points marked as CONTOURS
		myPoint.setComparator(myPoint.COMPARE_POINT_AND_TYPE);
		LinkedList<myPoint> queue = new LinkedList<myPoint>();
		queue.add(pxMin);
		while (!queue.isEmpty()) {
			myPoint p = queue.removeFirst();
			for(int i = -1; i <= 1; i += 2)
				for(int j = -1; j <= 1; j += 2) {
					int dx = (i + j) / 2, dy = (i - j) / 2;
					myPoint q = new myPoint(p.x + dx, p.y + dy, myPoint.BELT);
					if (belt.contains(q)) {
						belt.remove(q);
						q.type = myPoint.OUTSIDE_BELT;
						belt.add(q);
						queue.addLast(q);
					}
				}
		}
		
		//	other points are on the inside
		for(myPoint p : belt)
			if (p.type == myPoint.BELT)
				p.type = myPoint.INSIDE_BELT;
	}
}
